/******************************************************************************
 * $Id$
 * 
 * Source File Name: Cmapscope.c
 * 
 *   	 Copyright (c) 1991-96 International Business Machines Corp.
 *                     11400 Burnet Rd. Austin, TX. 78758
 *  
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of International Business Machines Corp. may not be used in 
 * any advertising or publicity relating to the software without the specific, 
 * prior written permission of International Business Machines Corp.
 * 
 * The software is provided AS-IS and without WARRANTY of any kind,
 * express, implied or otherwise, including without limitation, any
 * warranty of merchantabilityl or fitness for a particular purpose.
 * 
 * In no event shall International Business Machines Corp. be liable for any 
 * special, incidental, indirect or consequential damages of any kind, or any 
 * damages whatsoever resulting from loss of use, data or profits, whether or 
 * not advised of the possibility of damage, and on any theory of liability, 
 * arising out of or in connection with the use or performance of this 
 * software. Contractor is International Business Machines Corp.
 * 
 * Language: "C"
 * 
 * Description: 
 * 
 * Revision History: 
 *  01/07/97  Ronald O. Edmark  Initial Coding.
 *     
 *****************************************************************************/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#include <Xm/MwmUtil.h>
#include <Xm/Protocols.h>
#include <X11/cursorfont.h>
#include "cmapscope_ui.h"
#include "Cmapscope.h"

void dtb_save_toplevel_widget(Widget);

Widget dtb_get_toplevel_widget();

void dtb_save_command(char *);

static int determine_exe_dir_from_argv(
    char 	*argv0,
    char 	*buf,
    int 	bufSize
);

static int determine_exe_dir_from_path (
    char 	*argv0,
    char 	*buf,
    int 	bufSize
);

static Boolean path_is_executable(
    char 	*path,
    uid_t	euid,
    gid_t 	egid
);

int dtb_cmapscopeui_cmapscopewindow_initialize(
    DtbCmapscopeuiCmapscopewindowInfo	instance,
    Widget	parent
);

#include <X11/Xmu/Editres.h>

#ifndef min
#define min(a,b) ((a) < (b)? (a):(b))
#endif

#ifndef max
#define max(a,b) ((a) > (b)? (a):(b))
#endif

#ifndef ABS
#define ABS(x) ((x) >= 0? (x):(-(x)))
#endif

extern void cmsDecendXTree(Window);
extern void cmsChangeMap(Bool);
extern void  cmsColormapNotify(XColormapEvent *);

static Display *dtb_display;
static Screen *dtb_screen;

Display *dtb_get_display() {
  return dtb_display;
}

/*
 * Variable for storing top level widget
 */
static Widget		dtb_project_toplevel_widget = (Widget)NULL;

/*
 * Variable for storing command used to invoke application
 */
static char				*dtb_save_command_str = (char *)NULL;

/*
 * Directory where the binary for this process whate loaded from
 */
static char				*dtb_exe_dir = (char *)NULL;

/*
 * command line options...
 */
static XrmOptionDescRec optionDescList[] = {
    {"-session", "*session", XrmoptionSepArg, (XPointer)NULL}
    
};

/*
 * Application Resources
 */
static XtResource resources[] = {
    {"session", "Session", XtRString, sizeof(String),
        XtOffsetOf(DtbAppResourceRec, session_file), XtRImmediate, (XtPointer)NULL}
};

DtbAppResourceRec	dtb_app_resource_rec;

Boolean dtb_get_window_tree(XtPointer widget)
{
  XDefineCursor(XtDisplay((Widget)widget), XtWindow((Widget)widget), 
		XCreateFontCursor(XtDisplay((Widget)widget), XC_watch));
  cmsDecendXTree((Window)RootWindowOfScreen(dtb_screen));
  cmsChangeMap(True);
  XDefineCursor(XtDisplay((Widget)widget), XtWindow((Widget)widget), 
		XCreateFontCursor(XtDisplay((Widget)widget), XC_left_ptr));

  return True;
}


/*
 * main for application Cmapscope
 */
int
main(int argc, char **argv)
{
    Widget		toplevel = NULL;
    Display		*display = NULL;
    XtAppContext	app;
    Atom		save_yourself_atom;
    XEvent              event;
    Widget              install, uninstall;

    toplevel = XtVaAppInitialize(&app, "Cmapscope",
        optionDescList, XtNumber(optionDescList),
        &argc, argv, (String *)NULL,
        (ArgList) NULL, 0);

    /*
     * Get display and verify initialization was successful.
     */
    if (toplevel != NULL)
    {
        display = XtDisplayOfObject(toplevel);
    }
    if (display == NULL)
    {
        fprintf(stderr, "Could not open display.");
        exit(1);
    }
    
    /*
     * Save the toplevel widget so it can be fetched later as needed.
     */
    dtb_save_toplevel_widget(toplevel);
    
    /*
     * Save the command used to invoke the application.
     */
    dtb_save_command(argv[0]);
    
    XtGetApplicationResources(toplevel, (XtPointer)&dtb_app_resource_rec,
    	resources, XtNumber(resources), (Arg *)NULL, 0);
    
    dtb_display = display;
    dtb_screen = XtScreenOfObject(toplevel);
    XtAddEventHandler(toplevel, (EventMask)0, True,
		      (XtEventHandler)_XEditResCheckMessages, 0);

    /*
     * Initialize all global variables.
     */
    dtbCmapscopeuiCmapscopewindowInfo_clear(&dtb_cmapscopeui_cmapscopewindow);
    
    /*
     * Set up the application's root window.
     */
    dtb_cmapscopeui_cmapscopewindow.cmapscopewindow = toplevel;
    XtVaSetValues(dtb_cmapscopeui_cmapscopewindow.cmapscopewindow,
        XmNallowShellResize, TRUE,
        XmNtitle, "Cmapscope Window",
        NULL);
    
    dtb_cmapscopeui_cmapscopewindow_initialize(&(dtb_cmapscopeui_cmapscopewindow), dtb_get_toplevel_widget());
    
    /*
     * Map any initially-visible windows
     */
    
    save_yourself_atom = XmInternAtom(XtDisplay(toplevel),
    	"WM_SAVE_YOURSELF", False);
    
    XtAppAddWorkProc(app, dtb_get_window_tree, toplevel);
    XtRealizeWidget(toplevel);

    /*
     * Enter event loop
     */

    while(1) {
      XtAppNextEvent(app, &event);
      XtDispatchEvent(&event);
      switch (event.type) {
      case ColormapNotify:
	if ((event.xcolormap.state == ColormapInstalled) ||
	    (event.xcolormap.state == ColormapUninstalled)) {
#ifdef DEBUG
	  printf("type = %d\n serial = %ld\n send_event = %d\n window = %x\n colormap = %x\n new = %d\n state = %d\n",
		 event.xcolormap.type,
		 event.xcolormap.serial,
		 event.xcolormap.send_event,
		 event.xcolormap.window,
		 event.xcolormap.colormap,
		 event.xcolormap.new,
		 event.xcolormap.state);
#endif
	  cmsColormapNotify((XColormapEvent *) &(event.xcolormap));
	}
	break;
      }
    }

    /*    XtAppMainLoop(app); */
    exit(0);
}

void
dtb_save_toplevel_widget(
    Widget	toplevel
)
{
	dtb_project_toplevel_widget = toplevel;
}

Widget
dtb_get_toplevel_widget()
{
	return(dtb_project_toplevel_widget);
}

/*
 ** Routines to save and access the command used to invoke the application. 
 */
void
dtb_save_command(
    char	*argv0
)
{
    char	exe_dir[MAXPATHLEN+1];
    dtb_save_command_str = argv0; 

    /*
     * Save the path to the executable
     */
    if (determine_exe_dir(argv0, exe_dir, MAXPATHLEN+1) >= 0)
    {
	dtb_exe_dir = (char *)malloc(strlen(exe_dir)+1);
	if (dtb_exe_dir != NULL)
	{
	    strcpy(dtb_exe_dir, exe_dir);
	}
    }
}

unsigned long
dtb_cvt_resource_from_string(
    Widget		parent,
    String		res_type,
    unsigned int	size_of_type,
    String		res_str_value,
    unsigned long	error_value
)
{
    unsigned long	cvt_value_return = error_value;
    unsigned char	cvt_value1 = 0;
    unsigned short	cvt_value2 = 0;
    unsigned int	cvt_value3 = 0;
    unsigned long	cvt_value4 = 0;
    XtPointer		cvt_value_ptr = NULL;
    int			which_cvt_value = -1;
    XrmValue		source;
    XrmValue		dest;

    if (size_of_type > sizeof(cvt_value_return))
    {
	/* Type we are converting to is too large */
	return cvt_value_return;
    }

    /*
     * Get a data object of the appropriate size
     */
    if (size_of_type == sizeof(cvt_value1))
    {
	which_cvt_value = 1;
	cvt_value_ptr = (XtPointer)&cvt_value1;
    }
    else if (size_of_type == sizeof(cvt_value2))
    {
	which_cvt_value = 2;
	cvt_value_ptr = (XtPointer)&cvt_value2;
    }
    else if (size_of_type == sizeof(cvt_value3))
    {
	which_cvt_value = 3;
	cvt_value_ptr = (XtPointer)&cvt_value3;
    }
    else if (size_of_type == sizeof(cvt_value4))
    {
	which_cvt_value = 4;
	cvt_value_ptr = (XtPointer)&cvt_value4;
    }
    else
    {
	return cvt_value_return;
    }

    /*
     * Actually do the conversion
     */
    source.size = strlen(res_str_value) + 1;
    source.addr = res_str_value;

    dest.size   = size_of_type;
    dest.addr   = (char *)cvt_value_ptr;

    if (XtConvertAndStore(parent, XtRString, &source,
		res_type, &dest) != 0)
    {
	switch (which_cvt_value)
	{
	    case 1:
		cvt_value_return = (unsigned long)cvt_value1;
	    break;

	    case 2:
		cvt_value_return = (unsigned long)cvt_value2;
	    break;

	    case 3:
		cvt_value_return = (unsigned long)cvt_value3;
	    break;

	    case 4:
		cvt_value_return = (unsigned long)cvt_value4;
	    break;
	}
    }

    return cvt_value_return;
}


/*
 * Returns the directory that the executable for this process was loaded 
 * from.
 */
String
dtb_get_exe_dir(void)
{
    return dtb_exe_dir;
}


/*
 * Determines the directory the executable for this process was loaded from.
 */
static int 
determine_exe_dir(
    char 	*argv0,
    char 	*buf,
    int 	bufSize
)
{
    Boolean	foundDir= False;

    if ((buf == NULL) || (bufSize < 1))
    {
	return -1;
    }
    
    if (determine_exe_dir_from_argv(argv0, buf, bufSize) >= 0)
    {
	foundDir = True;
    }

    if (!foundDir)
    {
	if (determine_exe_dir_from_path(argv0, buf, bufSize) >= 0)
	{
	    foundDir = True;
	}
    }

    /*
     * Convert relative path to absolute, so that directory changes will
     * not affect us.
     */
    if (foundDir && (buf[0] != '/'))
    {
	char	cwd[MAXPATHLEN+1];
	char	*env_pwd = NULL;
	char	*path_prefix = NULL;
	char	abs_exe_dir[MAXPATHLEN+1];

	if (getcwd(cwd, MAXPATHLEN+1) != NULL)
	{
	    path_prefix = cwd;
	}
	else
	{
	    env_pwd = getenv("PWD");
	    if (env_pwd != NULL)
	    {
		path_prefix = env_pwd;
	    }
	}

	if (path_prefix != NULL)
	{
	    strcpy(abs_exe_dir, path_prefix);
	    if (strcmp(buf, ".") != 0)
	    {
		strcat(abs_exe_dir, "/");
		strcat(abs_exe_dir, buf);
	    }
	    strcpy(buf, abs_exe_dir);
	}
    }

    return foundDir? 0:-1;
}
/*
 *  Looks for absolute path in arv[0].
 */
static int
determine_exe_dir_from_argv(
    char 	*argv0,
    char 	*buf,
    int 	bufSize
)
{
    int		i= 0;
    Boolean	foundit= False;

    for (i= strlen(argv0)-1; (i >= 0) && (argv0[i] != '/'); --i)
    {
    }

    if (i >= 0)
    {
	int	maxStringSize= min(i, bufSize);
	strncpy(buf, argv0, maxStringSize);
	buf[min(maxStringSize, bufSize-1)]= 0;
	foundit = True;
    }

    return foundit? 0:-1;
}


/*
 * Assumes: bufSize > 0
 */
static int
determine_exe_dir_from_path (
    char 	*argv0,
    char 	*buf,
    int 	bufSize
)
{
    Boolean	foundDir= False;
    Boolean	moreDirs= True;
    char	*szExeName= argv0;
    int		iExeNameLen= strlen(szExeName);
    char	*szTemp= NULL;
    char	szPathVar[MAXPATHLEN+1];
    int		iPathVarLen= 0;
    char	szCurrentPath[MAXPATHLEN+1];
    int		iCurrentPathLen= 0;
    int		iCurrentPathStart= 0;
    int		i = 0;
    uid_t	euid= geteuid();
    uid_t	egid= getegid();

    szTemp= getenv("PATH");
    if (szTemp == NULL)
    {
	moreDirs= False;
    }
    else
    {
	strncpy(szPathVar, szTemp, MAXPATHLEN);
	szPathVar[MAXPATHLEN]= 0;
	iPathVarLen= strlen(szPathVar);
    }

    while ((!foundDir) && (moreDirs))
    {
	/* find the current directory name */
	for (i= iCurrentPathStart; (i < iPathVarLen) && (szPathVar[i] != ':'); 
	    )
    	{
	    ++i;
	}
	iCurrentPathLen= i - iCurrentPathStart;
	if ((iCurrentPathLen + iExeNameLen + 2) > MAXPATHLEN)
	{
	    iCurrentPathLen= MAXPATHLEN - (iExeNameLen + 2);
	}

	/* create a possible path to the executable */
	strncpy(szCurrentPath, &szPathVar[iCurrentPathStart], iCurrentPathLen);
	szCurrentPath[iCurrentPathLen]= 0;
	strcat(szCurrentPath, "/");
	strcat(szCurrentPath, szExeName);

	/* see if the executable exists (and we can execute it) */
	if (path_is_executable(szCurrentPath, euid, egid))
	{
	    foundDir= True;
	}

	/* skip past the current directory name */
	if (!foundDir)
	{
	    iCurrentPathStart+= iCurrentPathLen;
	    while (   (iCurrentPathStart < iPathVarLen) 
		   && (szPathVar[iCurrentPathStart] != ':') )
	    {
		++iCurrentPathStart;	/* find : */
	    }
	    if (iCurrentPathStart < iPathVarLen) 
	    {
		++iCurrentPathStart;	/* skip : */
	    }
	    if (iCurrentPathStart >= iPathVarLen)
	    {
		moreDirs= False;
	    }
	}
    } /* while !foundDir */

    if (foundDir)
    {
	szCurrentPath[iCurrentPathLen]= 0;
	strncpy(buf, szCurrentPath, bufSize);
	buf[bufSize-1]= 0;
    }
    return foundDir? 0:-1;
}

char * 
dtb_get_command() 
{
    return(dtb_save_command_str); 
}
static Boolean
path_is_executable(
    char 	*path,
    uid_t	euid,
    gid_t 	egid
)
{
    Boolean	bExecutable= False;
    struct stat	sStat;

    if (stat(path, &sStat) == 0)
    {
	Boolean	bDetermined= False;

	if (!bDetermined)
	{
	    if (!S_ISREG(sStat.st_mode))
	    {
		/* not a regular file */
		bDetermined= True;
		bExecutable= False;
	    }
	}

	if (!bDetermined)
	{
	    if (   (euid == 0) 
	        && (   ((sStat.st_mode & S_IXOTH) != 0)
		    || ((sStat.st_mode & S_IXGRP) != 0)
		    || ((sStat.st_mode & S_IXUSR) != 0) )
	       )
	    {
		bDetermined= True;
		bExecutable= True;
	    }
	}

	if (!bDetermined)
	{
	    if (   (((sStat.st_mode & S_IXOTH) != 0)    )
		|| (((sStat.st_mode & S_IXGRP) != 0) && (sStat.st_gid == egid))
		|| (((sStat.st_mode & S_IXUSR) != 0) && (sStat.st_gid == euid))
	       )
	    {
		bDetermined= True;
	        bExecutable= True;
	    }
	}
    } /* if stat */

    return bExecutable;
}
