/******************************************************************************
 * $Id$
 * 
 * Source File Name: cmaps.c
 * 
 *   	 Copyright (c) 1991-96 International Business Machines Corp.
 *                     11400 Burnet Rd. Austin, TX. 78758
 *  
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of International Business Machines Corp. may not be used in 
 * any advertising or publicity relating to the software without the specific, 
 * prior written permission of International Business Machines Corp.
 * 
 * The software is provided AS-IS and without WARRANTY of any kind,
 * express, implied or otherwise, including without limitation, any
 * warranty of merchantabilityl or fitness for a particular purpose.
 * 
 * In no event shall International Business Machines Corp. be liable for any 
 * special, incidental, indirect or consequential damages of any kind, or any 
 * damages whatsoever resulting from loss of use, data or profits, whether or 
 * not advised of the possibility of damage, and on any theory of liability, 
 * arising out of or in connection with the use or performance of this 
 * software. Contractor is International Business Machines Corp.
 * 
 * Language: "C"
 * 
 * Description: 
 * 
 * Revision History: 
 *  01/07/97  Ronald O. Edmark  Initial Coding.
 *     
 *****************************************************************************/
#include "cmaps.h"

/************************************************************************
  Example code that is add to the GUI layer. I'm using a app work proc.

  cmsDecendXTree(RootWindowOfScreen(screen));
  cmsChangeMap(True);

  *******************************************************************/
static start_window;

/* Interface rountines to the GUI */
extern void dtb_update_dialog(char[][24]);
extern void dtb_create_list(char[][16], int);
extern Widget dtb_get_toplevel_widget();
extern Display* dtb_get_display();
extern Widget dtb_get_drawing_widget();
extern Bool dtb_install_toggle();
extern Bool dtb_uninstall_toggle();

/* Exported routines */
void cmsChangeMap(Bool);
void cmsDecendXTree(Window);

/* Local routines */
static void cmsRealDecendXTree(Window);
static void cmsBuildColormapLists(Window);
static int cmsFindColormap(Colormap);
static int cmsInsertCmap(CSCmapInfoRec*);
static void cmsCreateColormapDisplayWindow(CSCmapInfoRec*);
static void cmsMapColormapDisplayWindow(int);

#define hash(c) ((c) % TABLESIZE)

static void cmsBuildColormapLists(Window window)
{
  XWindowAttributes    window_attributes;
  XVisualInfo          vistemplate, *vinfo;
  sovVisualInfo        sovtemplate, *sov_info;
  int                  items_returned;
  CSCmapInfoRec        cmap_entry;
  Display*             display = dtb_get_display();
  int                  colormap;

  if (!XGetWindowAttributes(display, window, &window_attributes)) {
    printf("Error: XGetWindowAttributes failed.\n");
    exit(-2);
  }
  
  /* On need to track input only window */
  if (window_attributes.class == InputOnly) 
    return;

  sovtemplate.vinfo.visualid = vistemplate.visualid = 
    XVisualIDFromVisual(window_attributes.visual);
  vinfo = XGetVisualInfo(display, VisualIDMask, &vistemplate, 
			 &items_returned);
  sov_info = sovGetVisualInfo(display, VisualIDMask, &sovtemplate, 
			      &items_returned);

  assert(vinfo && sov_info);

  cmap_entry.cmap_id      = window_attributes.colormap;
  cmap_entry.cmap_cells   = window_attributes.visual->map_entries;
  cmap_entry.state        = (Bool)window_attributes.map_installed;
  cmap_entry.install_req  = 0L;
  cmap_entry.ref_cnt      = 0L;
  cmap_entry.depth        = window_attributes.depth;
  cmap_entry.visual_class = vinfo->class;
  cmap_entry.visual_type  = sov_info->layer;
  cmap_entry.visual       = window_attributes.visual;
  cmap_entry.window_class = window_attributes.class;

  colormap = cmsInsertCmap(&cmap_entry);

  if (colormap == NEW) {
    /* Elect to receive Colormap Notifies */
    XSelectInput(display, window, 
		 window_attributes.your_event_mask | ColormapChangeMask);
  }
  else if (colormap == FULL) {
    printf("Error: To many colormaps, out of hash table space.\n");
    exit(-3);
  }

  XFree(vinfo);
  if (sov_info) free((void*) sov_info);
}

static int cmsFindColormap(Colormap id)
{
  int try, h;

  try = h = hash(id);

  do {
    /* no entry found */
    if ((cmaps_table[try])->cmap_id == id) {
      return try;
    }
    else if (cmaps_table[try] == NULL) {
      return EMPTY;
    }
    try = (try - 1 + TABLESIZE) % TABLESIZE;
  } while (try != h);
  return NOTFOUND;
}

static int cmsInsertCmap(CSCmapInfoRec* cmap) 
{
  int             h, try;
  CSCmapInfoRec  *cmap_entry;

  try = h = hash(cmap->cmap_id);
  do {
    /* new entry */
    if (cmaps_table[try] == NULL) {
      cmap_entry = (CSCmapInfoRec *) malloc(sizeof(CSCmapInfoRec));

      cmap_entry->cmap_id      = cmap->cmap_id;
      cmap_entry->cmap_cells   = cmap->cmap_cells;
      cmap_entry->state        = cmap->state;
      cmap_entry->install_req  = cmap->install_req;
      cmap_entry->ref_cnt      = 1;
      cmap_entry->depth        = cmap->depth;
      cmap_entry->visual_class = cmap->visual_class;
      cmap_entry->visual_type  = cmap->visual_type;
      cmap_entry->visual       = cmap->visual;
      cmap_entry->window_class = cmap->window_class;
      cmap_entry->window       = NULL;
      cmap_entry->gc           = NULL;
      cmap_entry->notify       = FALSE;

      cmsCreateColormapDisplayWindow(cmap_entry);

      cmaps_table[try] = cmap_entry;
      cmaps_locations[total_cmaps++] = try;

      return NEW;
    }
    /* same one found */
    else if ((cmaps_table[try])->cmap_id == cmap->cmap_id) {
      (cmaps_table[try])->state = cmap->state;
      (cmaps_table[try])->ref_cnt++;
      return OLD;
    }
    try = (try - 1 + TABLESIZE) % TABLESIZE;
  } while (try != h);
  /* Our table must be full */
  return FULL;
}

static void cmsCreateColormapDisplayWindow(CSCmapInfoRec *cmap_entry)
{
  Display* display = dtb_get_display();
  XSetWindowAttributes window_attribs;

  window_attribs.colormap = cmap_entry->cmap_id;
  window_attribs.background_pixel = None;
  window_attribs.border_pixel = None;

  cmap_entry->window = 
    XCreateWindow(display, 
		  XtWindow(dtb_get_drawing_widget()),
		  1, 1, WIDTH, HEIGHT, 0, 
		  cmap_entry->depth, 
		  cmap_entry->window_class, 
		  cmap_entry->visual,
		  CWBorderPixel | CWBackPixel | 
		  CWColormap, &window_attribs);
  assert(cmap_entry->window);

  cmap_entry->gc = XCreateGC(display, cmap_entry->window, 0, NULL);
  assert(cmap_entry->gc);
}

static void cmsMapColormapDisplayWindow(int location)
{
  int      i;
  Display* display = dtb_get_display();
  Window   window = (cmaps_table[location])->window;
  int      cells = (cmaps_table[location])->cmap_cells;
  GC       gc = (cmaps_table[location])->gc;

  XMapRaised(display, window);
  XClearWindow(display, window);
  for (i = 0; i < cells; i++) {
    XSetForeground(display, gc, i);
    XFillRectangle(display, window, gc, (int)((i % CELL) * CELLW),
		   (int)((i / CELL) * CELLH), CELLW, CELLH);
  }
  XFlush(display);
}

/* Exported Interface routines */
void cmsDecendXTree(Window window)
{
  static Window window_list[TABLESIZE];
  static char   cmap_ids[TABLESIZE][16];
  static Bool   inited = FALSE;
  int           i;

  if (!inited) {
    for (i = 0; i < TABLESIZE; i++) {
      cmaps_table[i] = (CSCmapInfoRec *)NULL;
      cmaps_locations[i] = -1;
    }
    total_cmaps = current_cmap = 0;
    inited = TRUE;
  }
  cmsRealDecendXTree( window);

  for (i = 0; i < total_cmaps; i++) {
    sprintf(cmap_ids[i], "%3d 0x%x", i+1,
	    (cmaps_table[cmaps_locations[i]])->cmap_id);
    window_list[i] = (cmaps_table[cmaps_locations[i]])->window;
  }

  dtb_create_list(cmap_ids, total_cmaps);
  
  XSetWMColormapWindows(dtb_get_display(), XtWindow(dtb_get_toplevel_widget()),
			window_list, total_cmaps);
}

static void cmsRealDecendXTree(Window window)
{
  Window root_window, parent_window, *children_windows = NULL;
  unsigned int children;
  int i;

  if (!XQueryTree(dtb_get_display(), window, &root_window, 
		  &parent_window, &children_windows, &children)) {
    printf("Error: XQueryTree, exiting program.\n");
    exit(-1);
  }

  cmsBuildColormapLists(window);

  /* The chlidren are returned with the bottom most window listed first and
     the topmost window listed last. */
  for (i = ((int)children) - 1; i >= 0; i--) {
    cmsRealDecendXTree(children_windows[i]);
  }

  XFree((char *)children_windows);
}

void cmsChangeMap(Bool do_both)
{
  static int location = -1;
  char cmaps[8][24];

  if (do_both) {
    location = cmaps_locations[current_cmap];
    sprintf(cmaps[0], "%d", current_cmap + 1);
    sprintf(cmaps[1], "%d", (cmaps_table[location])->depth);
    sprintf(cmaps[2], "%d", (cmaps_table[location])->ref_cnt);
    sprintf(cmaps[3], "%d", (cmaps_table[location])->install_req);
    sprintf(cmaps[4], "0x%x", (cmaps_table[location])->cmap_id);
    sprintf(cmaps[5], "%s",
	    CSXVisualClassNames[(cmaps_table[location])->visual_class]);
    sprintf(cmaps[6], "%s",
	    CSXVisualTypes[(cmaps_table[location])->visual_type]);
    sprintf(cmaps[7], "%s",
	    ((cmaps_table[location])->state) ? 
	    "Installed" : "Uninstalled");

    current_cmap = (++current_cmap) % total_cmaps;

    /* Interface to GUI */
    dtb_update_dialog(cmaps);

    cmsMapColormapDisplayWindow(location);
  }
  else if (!do_both && (location != -1)){
    cmsMapColormapDisplayWindow(location);
  }
}

void cmsColormapNotify(XColormapEvent *cmap_event)
{
  int try;

  /* This means that the colormap has gone away. */
  if (cmap_event->colormap == 0 && cmap_event->new)
    return;

  try = cmsFindColormap(cmap_event->colormap);

  if (try != EMPTY && try != NOTFOUND) {
    if (cmap_event->state == ColormapInstalled) {
      if (dtb_install_toggle() && (cmaps_table[try])->notify)
	printf("Install:   Colormap id=0x%x, for window=0x%x\n", 
	       cmap_event->colormap, cmap_event->window);

      /* We need to update out state flag */
      if((cmap_event->colormap) == (cmaps_table[try]->cmap_id)) { 
	(cmaps_table[try])->state =  True;
	(cmaps_table[try])->install_req += 1;
      }
    }
    else if (cmap_event->state == ColormapUninstalled) {
      if (dtb_uninstall_toggle() && (cmaps_table[try])->notify)
	printf("Uninstall: Colormap id=0x%x, for window=0x%x\n", 
	       cmap_event->colormap, cmap_event->window);

      if((cmap_event->colormap) == (cmaps_table[try]->cmap_id)) { 
	(cmaps_table[try])->state = False;	
      }

    }
  }
}

void cmsSelectColormap(int *location, int count)
{
  int i, position, k;

  /* Because we don't get deselects we must reset the notify flags back to
     false. */ 
  for (i = 0; i < total_cmaps; i++) {
    (cmaps_table[cmaps_locations[i]])->notify = False;
  }

  /* If there are some selected, then set the flags to true. */
  if (count > 0) {
    for (i = 0; i < count; i++) {
      position = location[i] - 1;
      k = cmaps_locations[position];
      if (k != -1)
	(cmaps_table[k])->notify = True;
    }
  }
}
